@@ -9,11 +9,12 @@ namespace Kakoune
 
 static FaceRegistry::FaceSpec parse_face(StringView facedesc)
 {
-    constexpr StringView invalid_face_error = "invalid face description, expected [<fg>][,<bg>][+<attr>][@base] or just [base]";
+    constexpr StringView invalid_face_error = "invalid face description, expected [<fg>][,<bg>[,<underline>]][+<attr>][@base] or just [base]";
     if (all_of(facedesc, [](char c){ return is_word(c); }) and not is_color_name(facedesc))
         return {Face{}, facedesc.str()};
 
     auto bg_it = find(facedesc, ',');
+    auto underline_it = bg_it == facedesc.end() ? bg_it : std::find(bg_it+1, facedesc.end(), ',');
     auto attr_it = find(facedesc, '+');
     auto base_it = find(facedesc, '@');
     if (bg_it != facedesc.end()
@@ -25,12 +26,19 @@ static FaceRegistry::FaceSpec parse_face(StringView facedesc)
 
     auto colors_end = std::min(attr_it, base_it);
 
+    auto parse_color = [](StringView spec) {
+        return spec.empty() ? Color::Default : str_to_color(spec);
+    };
+
     FaceRegistry::FaceSpec spec;
     auto& face = spec.face;
-    face.fg = colors_end != facedesc.begin() ?
-        str_to_color({facedesc.begin(), std::min(bg_it, colors_end)}) : Color::Default;
+    face.fg = parse_color({facedesc.begin(), std::min(bg_it, colors_end)});
     if (bg_it != facedesc.end())
-        face.bg = bg_it+1 != attr_it ? str_to_color({bg_it+1, colors_end}) : Color::Default;
+    {
+        face.bg = parse_color({bg_it+1, std::min(underline_it, colors_end)});
+        if (underline_it != facedesc.end())
+            face.underline = parse_color({underline_it+1, colors_end});
+    }
     if (attr_it != facedesc.end())
     {
         for (++attr_it; attr_it != base_it; ++attr_it)
@@ -38,6 +46,7 @@ static FaceRegistry::FaceSpec parse_face(StringView facedesc)
             switch (*attr_it)
             {
                 case 'u': face.attributes |= Attribute::Underline; break;
+                case 'c': face.attributes |= Attribute::CurlyUnderline; break;
                 case 'r': face.attributes |= Attribute::Reverse; break;
                 case 'b': face.attributes |= Attribute::Bold; break;
                 case 'B': face.attributes |= Attribute::Blink; break;
@@ -65,11 +74,13 @@ String to_string(Attribute attributes)
     struct Attr { Attribute attr; StringView name; }
     attrs[] {
         { Attribute::Underline, "u" },
+        { Attribute::CurlyUnderline, "c" },
         { Attribute::Reverse, "r" },
         { Attribute::Blink, "B" },
         { Attribute::Bold, "b" },
         { Attribute::Dim, "d" },
         { Attribute::Italic, "i" },
+        { Attribute::Strikethrough, "s" },
         { Attribute::Final, "F" },
         { Attribute::FinalFg, "f" },
         { Attribute::FinalBg, "g" },
@@ -89,7 +100,7 @@ String to_string(Attribute attributes)
 
 String to_string(Face face)
 {
-    return format("{},{}{}", face.fg, face.bg, face.attributes);
+    return format("{},{},{}{}", face.fg, face.bg, face.underline, face.attributes);
 }
 
 Face FaceRegistry::operator[](StringView facedesc) const
@@ -173,6 +184,8 @@ FaceRegistry::FaceRegistry()
         { "MenuInfo", {Face{ Color::Cyan, Color::Default }} },
         { "Information", {Face{ Color::Black, Color::Yellow }} },
         { "Error", {Face{ Color::Black, Color::Red }} },
+        { "DiagnosticError", {Face{ Color::Red, Color::Default }} },
+        { "DiagnosticWarning", {Face{ Color::Yellow, Color::Default }} },
         { "StatusLine", {Face{ Color::Cyan, Color::Default }} },
         { "StatusLineMode", {Face{ Color::Yellow, Color::Default }} },
         { "StatusLineInfo", {Face{ Color::Blue, Color::Default }} },
