@@ -10,6 +10,7 @@
 #include "string.hh"
 #include "unicode.hh"
 
+#include <limits>
 #include <cerrno>
 #include <cstdlib>
 #include <cstdio>
@@ -217,8 +218,6 @@ MappedFile::MappedFile(StringView filename)
 
     if (st.st_size == 0)
         return;
-    else if (st.st_size > std::numeric_limits<int>::max())
-        throw runtime_error("file is too big");
 
     data = (const char*)mmap(nullptr, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
     if (data == MAP_FAILED)
@@ -237,6 +236,8 @@ MappedFile::~MappedFile()
 
 MappedFile::operator StringView() const
 {
+    if (st.st_size > std::numeric_limits<int>::max())
+        throw runtime_error("file is too big");
     return { data, (int)st.st_size };
 }
 
@@ -246,6 +247,13 @@ bool file_exists(StringView filename)
     return stat(filename.zstr(), &st) == 0;
 }
 
+bool regular_file_exists(StringView filename)
+{
+    struct stat st;
+    return stat(filename.zstr(), &st) == 0 and
+           (st.st_mode & S_IFMT) == S_IFREG;
+}
+
 void write(int fd, StringView data)
 {
     const char* ptr = data.data();
@@ -271,43 +279,6 @@ void write_to_file(StringView filename, StringView data)
     write(fd, data);
 }
 
-struct BufferedWriter
-{
-    BufferedWriter(int fd)
-      : m_fd{fd}, m_exception_count{std::uncaught_exceptions()} {}
-
-    ~BufferedWriter() noexcept(false)
-    {
-        if (m_pos != 0 and m_exception_count == std::uncaught_exceptions())
-            Kakoune::write(m_fd, {m_buffer, m_pos});
-    }
-
-    void write(StringView data)
-    {
-        while (not data.empty())
-        {
-            const ByteCount length = data.length();
-            const ByteCount write_len = std::min(length, size - m_pos);
-            memcpy(m_buffer + (int)m_pos, data.data(), (int)write_len);
-            m_pos += write_len;
-            if (m_pos == size)
-            {
-                Kakoune::write(m_fd, {m_buffer, size});
-                m_pos = 0;
-            }
-            data = data.substr(write_len);
-        }
-    }
-
-private:
-    static constexpr ByteCount size = 4096;
-    int m_fd;
-    int m_exception_count;
-    ByteCount m_pos = 0;
-    char m_buffer[(int)size];
-};
-
-
 void write_buffer_to_fd(Buffer& buffer, int fd)
 {
     auto eolformat = buffer.options()["eolformat"].get<EolFormat>();
@@ -360,7 +331,8 @@ void write_buffer_to_file(Buffer& buffer, StringView filename,
     bool replace = method == WriteMethod::Replace;
     bool force = flags & WriteFlags::Force;
 
-    if ((replace or force) and ::stat(zfilename, &st) != 0)
+    if ((replace or force) and (::stat(zfilename, &st) != 0 or
+                                (st.st_mode & S_IFMT) != S_IFREG))
     {
         force = false;
         replace = false;
@@ -511,14 +483,20 @@ CandidateList complete_filename(StringView prefix, const Regex& ignored_regex,
     auto [dirname, fileprefix] = split_path(prefix);
     auto parsed_dirname = parse_filename(dirname);
 
-    const bool check_ignored_regex = not ignored_regex.empty() and
-        not regex_match(fileprefix.begin(), fileprefix.end(), ignored_regex);
+    Optional<ThreadedRegexVM<const char*, RegexMode::Forward | RegexMode::AnyMatch | RegexMode::NoSaves>> vm;
+    if (not ignored_regex.empty())
+    {
+        vm.emplace(*ignored_regex.impl());
+        if (vm->exec(fileprefix.begin(), fileprefix.end(), fileprefix.begin(), fileprefix.end(), RegexExecFlags::None))
+            vm.reset();
+    }
+
     const bool only_dirs = (flags & FilenameFlags::OnlyDirectories);
 
-    auto filter = [&ignored_regex, check_ignored_regex, only_dirs](const dirent& entry, struct stat& st)
+    auto filter = [&vm, only_dirs](const dirent& entry, struct stat& st)
     {
         StringView name{entry.d_name};
-        return (not check_ignored_regex or not regex_match(name.begin(), name.end(), ignored_regex)) and
+        return (not vm or not vm->exec(name.begin(), name.end(), name.begin(), name.end(), RegexExecFlags::None)) and
                (not only_dirs or S_ISDIR(st.st_mode));
     };
     auto files = list_files(parsed_dirname, filter);
@@ -619,7 +597,7 @@ FsStatus get_fs_status(StringView filename)
 String get_kak_binary_path()
 {
     char buffer[2048];
-#if defined(__linux__) or defined(__CYGWIN__)
+#if defined(__linux__) or defined(__CYGWIN__) or defined(__gnu_hurd__)
     ssize_t res = readlink("/proc/self/exe", buffer, 2048);
     kak_assert(res != -1);
     buffer[res] = '\0';
@@ -654,6 +632,11 @@ String get_kak_binary_path()
     return buffer;
 #elif defined(__OpenBSD__)
     return KAK_BIN_PATH;
+#elif defined(__sun__)
+    ssize_t res = readlink("/proc/self/path/a.out", buffer, 2048);
+    kak_assert(res != -1);
+    buffer[res] = '\0';
+    return buffer;
 #else
 # error "finding executable path is not implemented on this platform"
 #endif
