@@ -86,9 +86,12 @@ InsertCompletion complete_word(const SelectionList& sels,
     HashMap<StringView, int> sel_word_counts;
     for (int i = 0; i < sels.size(); ++i)
     {
+        int len = 0;
+        auto is_short_enough_word = [&] (Codepoint c) { return len++ < WordDB::max_word_len && is_word_pred(c); };
+
         Utf8It end{buffer.iterator_at(sels[i].cursor()), buffer};
         Utf8It begin = end-1;
-        if (not skip_while_reverse(begin, buffer.begin(), is_word_pred) and
+        if (not skip_while_reverse(begin, buffer.begin(), is_short_enough_word) and
             begin < end) // (begin might == end if end == buffer.begin())
             ++begin;
 
@@ -98,10 +101,13 @@ InsertCompletion complete_word(const SelectionList& sels,
             prefix = buffer.substr(word_begin, end.base().coord());
         }
 
-        skip_while(end, buffer.end(), is_word_pred);
+        skip_while(end, buffer.end(), is_short_enough_word);
 
-        StringView word = buffer.substr(begin.base().coord(), end.base().coord());
-        ++sel_word_counts[word];
+        if (len <= WordDB::max_word_len)
+        {
+            StringView word = buffer.substr(begin.base().coord(), end.base().coord());
+            ++sel_word_counts[word];
+        }
     }
 
     struct RankedMatchAndBuffer : RankedMatch
@@ -261,76 +267,74 @@ InsertCompletion complete_option(const SelectionList& sels,
     auto& desc = opt.prefix;
     static const Regex re(R"((\d+)\.(\d+)(?:\+(\d+))?@(\d+))");
     MatchResults<String::const_iterator> match;
-    if (regex_match(desc.begin(), desc.end(), match, re))
+    if (not regex_match(desc.begin(), desc.end(), match, re))
+        return {};
+
+    BufferCoord coord{str_to_int({match[1].first, match[1].second}) - 1,
+                      str_to_int({match[2].first, match[2].second}) - 1};
+    if (not buffer.is_valid(coord))
+        return {};
+    size_t timestamp = (size_t)str_to_int({match[4].first, match[4].second});
+    auto changes = buffer.changes_since(timestamp);
+    if (any_of(changes, [&](auto&& change) { return change.begin < coord; }))
+        return {};
+
+    if (cursor_pos.line != coord.line or cursor_pos.column < coord.column)
+        return {};
+
+    const ColumnCount tabstop = options["tabstop"].get<int>();
+    const ColumnCount column = get_column(buffer, tabstop, cursor_pos);
+
+    struct RankedMatchAndInfo : RankedMatch
     {
-        BufferCoord coord{ str_to_int({match[1].first, match[1].second}) - 1,
-                         str_to_int({match[2].first, match[2].second}) - 1 };
-        if (not buffer.is_valid(coord))
-            return {};
-        size_t timestamp = (size_t)str_to_int({match[4].first, match[4].second});
-        auto changes = buffer.changes_since(timestamp);
-        if (any_of(changes, [&](auto&& change) { return change.begin < coord; }))
-            return {};
+        using RankedMatch::RankedMatch;
+        using RankedMatch::operator==;
+        using RankedMatch::operator<;
 
-        if (cursor_pos.line == coord.line and cursor_pos.column >= coord.column)
+        StringView on_select;
+        DisplayLine menu_entry;
+    };
+
+    StringView query = buffer.substr(coord, cursor_pos);
+    Vector<RankedMatchAndInfo> matches;
+
+    for (auto& candidate : opt.list)
+    {
+        if (RankedMatchAndInfo match{std::get<0>(candidate), query})
         {
-            StringView query = buffer.substr(coord, cursor_pos);
+            match.on_select = std::get<1>(candidate);
+            auto& menu = std::get<2>(candidate);
+            match.menu_entry = not menu.empty() ?
+                parse_display_line(expand_tabs(menu, tabstop, column), faces)
+              : DisplayLine{String{}, {}};
 
-            const ColumnCount tabstop = options["tabstop"].get<int>();
-            const ColumnCount column = get_column(buffer, tabstop, cursor_pos);
-
-            struct RankedMatchAndInfo : RankedMatch
-            {
-                using RankedMatch::RankedMatch;
-                using RankedMatch::operator==;
-                using RankedMatch::operator<;
-
-                StringView on_select;
-                DisplayLine menu_entry;
-            };
-
-            Vector<RankedMatchAndInfo> matches;
-
-            for (auto& candidate : opt.list)
-            {
-                if (RankedMatchAndInfo match{std::get<0>(candidate), query})
-                {
-                    match.on_select = std::get<1>(candidate);
-                    auto& menu = std::get<2>(candidate);
-                    match.menu_entry = not menu.empty() ?
-                        parse_display_line(expand_tabs(menu, tabstop, column), faces)
-                      : DisplayLine{ expand_tabs(menu, tabstop, column), {} };
-
-                    matches.push_back(std::move(match));
-                }
-            }
-
-            constexpr size_t max_count = 100;
-            // Gather best max_count matches
-            auto greater = [](auto& lhs, auto& rhs) { return rhs < lhs; };
-            auto first = matches.begin(), last = matches.end();
-            std::make_heap(first, last, greater);
-            InsertCompletion::CandidateList candidates;
-            candidates.reserve(std::min(matches.size(), max_count));
-            candidates.reserve(matches.size());
-            while (candidates.size() < max_count and first != last)
-            {
-                if (candidates.empty() or candidates.back().completion != first->candidate())
-                    candidates.push_back({ first->candidate().str(), first->on_select.str(),
-                                           std::move(first->menu_entry) });
-                std::pop_heap(first, last--, greater);
-            }
-
-            auto end = cursor_pos;
-            if (match[3].matched)
-            {
-                ByteCount len = str_to_int({match[3].first, match[3].second});
-                end = buffer.advance(coord, len);
-            }
-            return { std::move(candidates), coord, end, timestamp };
+            matches.push_back(std::move(match));
         }
     }
-    return {};
+
+    constexpr size_t max_count = 100;
+    // Gather best max_count matches
+    auto greater = [](auto& lhs, auto& rhs) { return rhs < lhs; };
+    auto first = matches.begin(), last = matches.end();
+    std::make_heap(first, last, greater);
+    InsertCompletion::CandidateList candidates;
+    candidates.reserve(std::min(matches.size(), max_count));
+    candidates.reserve(matches.size());
+    while (candidates.size() < max_count and first != last)
+    {
+        if (candidates.empty() or candidates.back().completion != first->candidate())
+            candidates.push_back({ first->candidate().str(), first->on_select.str(),
+                                   std::move(first->menu_entry) });
+        std::pop_heap(first, last--, greater);
+    }
+
+    auto end = cursor_pos;
+    if (match[3].matched)
+    {
+        ByteCount len = str_to_int({match[3].first, match[3].second});
+        end = buffer.advance(coord, len);
+    }
+    return { std::move(candidates), coord, end, timestamp };
 }
 
 template<bool other_buffers>
@@ -344,7 +348,8 @@ InsertCompletion complete_line(const SelectionList& sels,
     const ColumnCount tabstop = options["tabstop"].get<int>();
     const ColumnCount column = get_column(buffer, tabstop, cursor_pos);
 
-    StringView prefix = buffer[cursor_pos.line].substr(0_byte, cursor_pos.column);
+    String prefix = trim_indent(buffer[cursor_pos.line].substr(0_byte, cursor_pos.column));
+    BufferCoord replace_begin = buffer.advance(cursor_pos, -prefix.length());
     InsertCompletion::CandidateList candidates;
 
     auto add_candidates = [&](const Buffer& buf) {
@@ -352,12 +357,16 @@ InsertCompletion complete_line(const SelectionList& sels,
         {
             if (buf.name() == buffer.name() && l == cursor_pos.line)
                 continue;
-            const StringView line = buf[l];
+
+            String line = trim_indent(buf[l]);
+
+            if (line.length() == 0)
+              continue;
+
             if (prefix == line.substr(0_byte, prefix.length()))
             {
-                StringView candidate = line.substr(0_byte, line.length()-1);
-                candidates.push_back({candidate.str(), "",
-                                      {expand_tabs(candidate, tabstop, column), {}}});
+                String candidate = trim_indent(line.substr(0_byte, line.length()));
+                candidates.push_back({candidate, "", {expand_tabs(candidate, tabstop, column), {}} });
                 // perf: it's unlikely the user intends to search among >10 candidates anyway
                 if (candidates.size() == 100)
                     break;
@@ -380,7 +389,7 @@ InsertCompletion complete_line(const SelectionList& sels,
         return {};
     std::sort(candidates.begin(), candidates.end());
     candidates.erase(std::unique(candidates.begin(), candidates.end()), candidates.end());
-    return { std::move(candidates), cursor_pos.line, cursor_pos, buffer.timestamp() };
+    return { std::move(candidates), replace_begin, cursor_pos, buffer.timestamp() };
 }
 
 }
@@ -412,7 +421,6 @@ void InsertCompleter::select(int index, bool relative, Vector<Key>& keystrokes)
     const auto suffix_len = std::max(0_byte, buffer.distance(cursor_pos, m_completions.end));
 
     auto ref = buffer.string(m_completions.begin, m_completions.end);
-    ForwardChangesTracker changes_tracker;
     Vector<BufferRange> ranges;
     for (auto& sel : selections)
     {
