@@ -80,6 +80,8 @@ static constexpr KeyAndName keynamemap[] = {
     { "minus", '-' },
     { "semicolon", ';' },
     { "percent", '%' },
+    { "focus_in", Key::FocusIn },
+    { "focus_out", Key::FocusOut },
 };
 
 KeyList parse_keys(StringView str)
@@ -229,6 +231,10 @@ UnitTest test_keys{[]()
          alt('j'),
          ctrl('r'),
          shift(Key::Up),
+         ctrl('['),
+         ctrl('\\'),
+         ctrl(']'),
+         ctrl('_'),
     };
     String keys_as_str;
     for (auto& key : keys)
