@@ -201,8 +201,10 @@ void LineRangeSet::remove_range(LineRange range)
 
 UnitTest test_line_modifications{[]()
 {
+    auto make_lines = [](auto&&... lines) { return BufferLines{StringData::create({lines})...}; };
+
     {
-        Buffer buffer("test", Buffer::Flags::None, "line 1\nline 2\n");
+        Buffer buffer("test", Buffer::Flags::None, make_lines("line 1\n", "line 2\n"));
         auto ts = buffer.timestamp();
         buffer.erase({1, 0}, {2, 0});
 
@@ -211,7 +213,7 @@ UnitTest test_line_modifications{[]()
     }
 
     {
-        Buffer buffer("test", Buffer::Flags::None, "line 1\nline 2\n");
+        Buffer buffer("test", Buffer::Flags::None, make_lines("line 1\n", "line 2\n"));
         auto ts = buffer.timestamp();
         buffer.insert({2, 0}, "line 3");
 
@@ -220,7 +222,7 @@ UnitTest test_line_modifications{[]()
     }
 
     {
-        Buffer buffer("test", Buffer::Flags::None, "line 1\nline 2\nline 3\n");
+        Buffer buffer("test", Buffer::Flags::None, make_lines("line 1\n", "line 2\n", "line 3\n"));
 
         auto ts = buffer.timestamp();
         buffer.insert({1, 4}, "hoho\nhehe");
@@ -231,7 +233,7 @@ UnitTest test_line_modifications{[]()
     }
 
     {
-        Buffer buffer("test", Buffer::Flags::None, "line 1\nline 2\nline 3\nline 4\n");
+        Buffer buffer("test", Buffer::Flags::None, make_lines("line 1\n", "line 2\n", "line 3\n", "line 4\n"));
 
         auto ts = buffer.timestamp();
         buffer.erase({0,0}, {3,0});
@@ -250,7 +252,7 @@ UnitTest test_line_modifications{[]()
     }
 
     {
-        Buffer buffer("test", Buffer::Flags::None, "line 1\n");
+        Buffer buffer("test", Buffer::Flags::None, make_lines("line 1\n"));
         auto ts = buffer.timestamp();
         buffer.insert({0,0}, "n");
         buffer.insert({0,1}, "e");
