@@ -12,6 +12,7 @@
 
 #include <cstdio>
 #include <cstring>
+#include <limits>
 
 namespace Kakoune
 {
@@ -73,13 +74,14 @@ struct ParsedRegex
     };
 
     using NodeIndex = uint16_t;
-    struct Node
+    struct [[gnu::packed]] Node
     {
         Op op;
         bool ignore_case;
         NodeIndex children_end;
         Codepoint value;
         Quantifier quantifier;
+        uint16_t filler = 0;
     };
     static_assert(sizeof(Node) == 16, "");
 
@@ -1240,17 +1242,14 @@ namespace
 template<RegexMode mode = RegexMode::Forward>
 struct TestVM : CompiledRegex, ThreadedRegexVM<const char*, mode>
 {
-    using VMType = ThreadedRegexVM<const char*, mode>;
-
-    TestVM(StringView re, bool dump = false)
-        : CompiledRegex{compile_regex(re, mode & RegexMode::Forward ?
-                                          RegexCompileFlags::None : RegexCompileFlags::Backward)},
-          VMType{(const CompiledRegex&)*this}
-    { if (dump) puts(dump_regex(*this).c_str()); }
+    TestVM(StringView re)
+      : CompiledRegex{compile_regex(re, mode & RegexMode::Forward ? RegexCompileFlags::None : RegexCompileFlags::Backward)},
+        TestVM::ThreadedRegexVM{static_cast<const CompiledRegex&>(*this)}
+    {}
 
     bool exec(StringView re, RegexExecFlags flags = RegexExecFlags::None)
     {
-        return VMType::exec(re.begin(), re.end(), re.begin(), re.end(), flags);
+        return TestVM::ThreadedRegexVM::exec(re.begin(), re.end(), re.begin(), re.end(), flags);
     }
 };
 }
