@@ -461,7 +461,7 @@ static bool send_data(int fd, RemoteBuffer& buffer, Optional<int> ancillary_fd =
 }
 
 RemoteUI::RemoteUI(int socket, DisplayCoord dimensions)
-    : m_socket_watcher(socket,  FdEvents::Read | FdEvents::Write,
+    : m_socket_watcher(socket,  FdEvents::Read | FdEvents::Write, EventMode::Urgent,
                        [this](FDWatcher& watcher, FdEvents events, EventMode) {
           const int sock = watcher.fd();
           try
@@ -587,24 +587,20 @@ String get_user_name()
     return getenv("USER");
 }
 
-String session_directory()
+const String& session_directory()
 {
-    StringView xdg_runtime_dir = getenv("XDG_RUNTIME_DIR");
-    if (xdg_runtime_dir.empty())
-        return format("{}/kakoune/{}", tmpdir(), get_user_name());
-    else
-        return format("{}/kakoune", xdg_runtime_dir);
-}
-
-void make_session_directory()
-{
-    StringView xdg_runtime_dir = getenv("XDG_RUNTIME_DIR");
-    if (xdg_runtime_dir.empty())
-    {
-        // set sticky bit on the shared kakoune directory
-        make_directory(format("{}/kakoune", tmpdir()), 01777);
-    }
-    make_directory(session_directory(), 0711);
+    static String session_dir = [] {
+        StringView xdg_runtime_dir = getenv("XDG_RUNTIME_DIR");
+        if (not xdg_runtime_dir.empty())
+        {
+            if (struct stat st; stat(xdg_runtime_dir.zstr(), &st) == 0 && st.st_uid == geteuid())
+                return format("{}/kakoune", xdg_runtime_dir);
+            else
+                write_to_debug_buffer("XDG_RUNTIME_DIR does not exist or not owned by current user, using tmpdir");
+        }
+        return format("{}/kakoune-{}", tmpdir(), get_user_name());
+    }();
+    return session_dir;
 }
 
 String session_path(StringView session)
@@ -659,7 +655,7 @@ RemoteClient::RemoteClient(StringView session, StringView name, std::unique_ptr<
         m_socket_watcher->events() |= FdEvents::Write;
      });
 
-    m_socket_watcher.reset(new FDWatcher{sock, FdEvents::Read | FdEvents::Write,
+    m_socket_watcher.reset(new FDWatcher{sock, FdEvents::Read | FdEvents::Write, EventMode::Urgent,
                            [this, reader = MsgReader{}](FDWatcher& watcher, FdEvents events, EventMode) mutable {
         const int sock = watcher.fd();
         if (events & FdEvents::Write and send_data(sock, m_send_buffer))
@@ -773,7 +769,7 @@ class Server::Accepter
 {
 public:
     Accepter(int socket)
-        : m_socket_watcher(socket, FdEvents::Read,
+        : m_socket_watcher(socket, FdEvents::Read, EventMode::Urgent,
                            [this](FDWatcher&, FdEvents, EventMode mode) {
                                handle_available_input(mode);
                            })
@@ -849,8 +845,8 @@ private:
     MsgReader m_reader;
 };
 
-Server::Server(String session_name)
-    : m_session{std::move(session_name)}
+Server::Server(String session_name, bool is_daemon)
+    : m_session{std::move(session_name)}, m_is_daemon{is_daemon}
 {
     if (not all_of(m_session, is_identifier))
         throw runtime_error{format("invalid session name: '{}'", session_name)};
@@ -859,7 +855,7 @@ Server::Server(String session_name)
     fcntl(listen_sock, F_SETFD, FD_CLOEXEC);
     sockaddr_un addr = session_addr(m_session);
 
-    make_session_directory();
+    make_directory(session_directory(), 0711);
 
     // Do not give any access to the socket to other users by default
     auto old_mask = umask(0077);
@@ -884,7 +880,7 @@ Server::Server(String session_name)
 
         m_accepters.emplace_back(new Accepter{sock});
     };
-    m_listener.reset(new FDWatcher{listen_sock, FdEvents::Read, accepter});
+    m_listener.reset(new FDWatcher{listen_sock, FdEvents::Read, EventMode::Urgent, accepter});
 }
 
 bool Server::rename_session(StringView name)
